@@ -1,3 +1,5 @@
+import time
+import logging
 import sys
 import argparse
 from statistics import mean
@@ -6,6 +8,7 @@ import binascii
 
 from humanfriendly import format_size
 from humanfriendly import parse_size
+from humanfriendly import format_timespan
 from elftools.elf.elffile import ELFFile
 
 from .create import create_patch
@@ -131,13 +134,7 @@ def data_format_from_files(option, elffile, binfile, offset):
             data_range.end)
 
 
-def _do_create_patch(args):
-    if args.type == 'in-place':
-        if args.memory_size is None:
-            raise Error('--memory-size is required for in-place patch.')
-        elif args.segment_size is None:
-            raise Error('--segment-size is required for in-place patch.')
-
+def data_format_args(args):
     from_data_offset_begin, from_data_offset_end = parse_range(
         '--from-data-offsets',
         args.from_data_offsets)
@@ -201,41 +198,87 @@ def _do_create_patch(args):
             to_data_begin,
             to_data_end))
 
+    return {
+        'data_format': args.data_format,
+        'from_data_offset_begin': from_data_offset_begin,
+        'from_data_offset_end': from_data_offset_end,
+        'from_data_begin': from_data_begin,
+        'from_data_end': from_data_end,
+        'from_code_begin': from_code_begin,
+        'from_code_end': from_code_end,
+        'to_data_offset_begin': to_data_offset_begin,
+        'to_data_offset_end': to_data_offset_end,
+        'to_data_begin': to_data_begin,
+        'to_data_end': to_data_end,
+        'to_code_begin': to_code_begin,
+        'to_code_end': to_code_end
+    }
+
+
+def print_successful(filename, start_time):
+    print("Successfully created '{}' in {}!".format(
+        filename,
+        format_timespan(time.time() - start_time)))
+
+
+def _do_create_patch(args):
+    start_time = time.time()
     create_patch_filenames(args.fromfile,
                            args.tofile,
                            args.patchfile,
                            args.compression,
-                           args.type,
+                           args.patch_type,
+                           args.algorithm,
+                           args.suffix_array_algorithm,
+                           match_score=args.match_score,
+                           match_block_size=args.match_block_size,
+                           use_mmap=not args.no_mmap,
+                           **data_format_args(args))
+    print_successful(args.patchfile, start_time)
+
+
+def _do_create_patch_in_place(args):
+    start_time = time.time()
+    create_patch_filenames(args.fromfile,
+                           args.tofile,
+                           args.patchfile,
+                           args.compression,
+                           'in-place',
+                           'bsdiff',
+                           args.suffix_array_algorithm,
                            args.memory_size,
                            args.segment_size,
                            args.minimum_shift_size,
-                           args.data_format,
-                           from_data_offset_begin,
-                           from_data_offset_end,
-                           from_data_begin,
-                           from_data_end,
-                           from_code_begin,
-                           from_code_end,
-                           to_data_offset_begin,
-                           to_data_offset_end,
-                           to_data_begin,
-                           to_data_end,
-                           to_code_begin,
-                           to_code_end)
+                           use_mmap=not args.no_mmap,
+                           **data_format_args(args))
+    print_successful(args.patchfile, start_time)
 
-    print("Successfully created patch '{}'!".format(args.patchfile))
+
+def _do_create_patch_bsdiff(args):
+    start_time = time.time()
+    create_patch_filenames(args.fromfile,
+                           args.tofile,
+                           args.patchfile,
+                           patch_type='bsdiff')
+    print_successful(args.patchfile, start_time)
 
 
 def _do_apply_patch(args):
+    start_time = time.time()
     apply_patch_filenames(args.fromfile, args.patchfile, args.tofile)
+    print_successful(args.tofile, start_time)
 
 
 def _do_apply_patch_in_place(args):
+    start_time = time.time()
     apply_patch_in_place_filenames(args.memfile, args.patchfile)
+    print_successful(args.memfile, start_time)
 
 
 def _do_apply_patch_bsdiff(args):
+    start_time = time.time()
     apply_patch_bsdiff_filenames(args.fromfile, args.patchfile, args.tofile)
+    print_successful(args.tofile, start_time)
 
 
 def _format_size(value):
@@ -302,18 +345,18 @@ def _patch_info_in_place_segment(fsize,
     print()
 
 
-def _patch_info_normal(detailed,
-                       fsize,
-                       patch_size,
-                       compression,
-                       dfpatch_size,
-                       data_format,
-                       dfpatch_info,
-                       to_size,
-                       diff_sizes,
-                       extra_sizes,
-                       adjustment_sizes,
-                       number_of_size_bytes):
+def _patch_info_sequential(detailed,
+                           fsize,
+                           patch_size,
+                           compression,
+                           dfpatch_size,
+                           data_format,
+                           dfpatch_info,
+                           to_size,
+                           diff_sizes,
+                           extra_sizes,
+                           adjustment_sizes,
+                           number_of_size_bytes):
     del adjustment_sizes
 
     number_of_diff_bytes = sum(diff_sizes)
@@ -337,7 +380,7 @@ def _patch_info_normal(detailed,
         mean_extra_size = '-'
         median_extra_size = '-'
 
-    print('Type:               normal')
+    print('Type:               sequential')
     print('Patch size:         {}'.format(fsize(patch_size)))
     print('To size:            {}'.format(fsize(to_size)))
     print('Patch/to ratio:     {} % (lower is better)'.format(patch_to_ratio))
@@ -390,7 +433,7 @@ def _patch_info_in_place(fsize,
     print('Compression:        {}'.format(compression))
     print()
 
-    for i, (dfpatch_size, data_format, normal_info) in enumerate(segments):
+    for i, (dfpatch_size, data_format, sequential_info) in enumerate(segments):
         from_offset_begin = max(segment_size * (i + 1) - from_shift_size, 0)
         from_offset_end = min(from_size, memory_size - from_shift_size)
         to_offset_begin = (segment_size * i)
@@ -403,7 +446,20 @@ def _patch_info_in_place(fsize,
                                      to_offset_end,
                                      dfpatch_size,
                                      data_format,
-                                     *normal_info)
+                                     *sequential_info)
+
+
+def _patch_info_hdiffpatch(fsize,
+                           patch_size,
+                           compression,
+                           to_size):
+    patch_to_ratio = _format_ratio(patch_size, to_size)
+
+    print('Type:               hdiffpatch')
+    print('Patch size:         {}'.format(fsize(patch_size)))
+    print('To size:            {}'.format(fsize(to_size)))
+    print('Patch/to ratio:     {} % (lower is better)'.format(patch_to_ratio))
+    print('Compression:        {}'.format(compression))
 
 
 def _do_patch_info(args):
@@ -414,10 +470,12 @@ def _do_patch_info(args):
 
     patch_type, info = patch_info_filename(args.patchfile, fsize)
 
-    if patch_type == 'normal':
-        _patch_info_normal(args.detailed, fsize, *info)
+    if patch_type == 'sequential':
+        _patch_info_sequential(args.detailed, fsize, *info)
     elif patch_type == 'in-place':
         _patch_info_in_place(fsize, *info)
+    elif patch_type == 'hdiffpatch':
+        _patch_info_hdiffpatch(fsize, *info)
     else:
         raise Error('Bad patch type {}.'.format(patch_type))
 
@@ -426,42 +484,7 @@ def to_binary_size(value):
     return parse_size(value, binary=True)
 
 
-def _main():
-    parser = argparse.ArgumentParser(description='Binary delta encoding utility.')
-
-    parser.add_argument('-d', '--debug', action='store_true')
-    parser.add_argument('--version',
-                        action='version',
-                        version=__version__,
-                        help='Print version information and exit.')
-
-    # Workaround to make the subparser required in Python 3.
-    subparsers = parser.add_subparsers(title='subcommands',
-                                       dest='subcommand')
-    subparsers.required = True
-
-    # Create patch subparser.
-    subparser = subparsers.add_parser('create_patch',
-                                      description='Create a patch.')
-    subparser.add_argument('-t', '--type',
-                           choices=('normal', 'in-place', 'bsdiff'),
-                           default='normal',
-                           help='Patch type (default: %(default)s).')
-    subparser.add_argument('-c', '--compression',
-                           choices=sorted(_COMPRESSIONS),
-                           default='lzma',
-                           help='Compression algorithm (default: %(default)s).')
-    subparser.add_argument('--memory-size',
-                           type=to_binary_size,
-                           help='Target memory size.')
-    subparser.add_argument(
-        '--segment-size',
-        type=to_binary_size,
-        help='Segment size. Must be a multiple of the largest erase block size.')
-    subparser.add_argument(
-        '--minimum-shift-size',
-        type=to_binary_size,
-        help='Minimum shift size (default: 2 * segment size).')
+def add_data_format_args(subparser):
     subparser.add_argument(
         '--data-format',
         choices=sorted(_DATA_FORMATS),
@@ -496,14 +519,118 @@ def _main():
     subparser.add_argument(
         '--to-data-addresses',
         help='To file data address ranges.')
+
+
+def _main():
+    parser = argparse.ArgumentParser(description='Binary delta encoding utility.')
+
+    parser.add_argument('-d', '--debug', action='store_true')
+    parser.add_argument('-l', '--log-level',
+                        default='error',
+                        choices=[
+                            'debug', 'info', 'warning', 'error', 'critical'
+                        ],
+                        help='Set the logging level (default: %(default)s).')
+    parser.add_argument('--version',
+                        action='version',
+                        version=__version__,
+                        help='Print version information and exit.')
+
+    # Workaround to make the subparser required in Python 3.
+    subparsers = parser.add_subparsers(title='subcommands',
+                                       dest='subcommand')
+    subparsers.required = True
+
+    # Create sequential patch subparser.
+    subparser = subparsers.add_parser('create_patch',
+                                      description='Create a sequential patch.')
+    subparser.add_argument(
+        '-c', '--compression',
+        choices=sorted(_COMPRESSIONS),
+        default='lzma',
+        help='Compression algorithm (default: %(default)s).')
+    subparser.add_argument(
+        '-t', '--patch-type',
+        choices=('sequential', 'hdiffpatch'),
+        default='sequential',
+        help=('Patch type. Sequential is sequential, hdiffpatch smaller '
+              '(default: %(default)s).'))
+    subparser.add_argument(
+        '-a', '--algorithm',
+        choices=('bsdiff', 'hdiffpatch', 'match-blocks'),
+        default='bsdiff',
+        help='Diff algorithm (default: %(default)s).')
+    subparser.add_argument(
+        '-s', '--suffix-array-algorithm',
+        choices=('sais', 'divsufsort'),
+        default='divsufsort',
+        help=('Suffix array algorithm used by bsdiff algorithm '
+              '(default: %(default)s).'))
+    subparser.add_argument(
+        '--match-score',
+        type=int,
+        default=6,
+        help='Match score used by hdiffpatch algorithm (default: %(default)s).')
+    subparser.add_argument(
+        '--match-block-size',
+        type=to_binary_size,
+        default=64,
+        help=('Match block size used by match-blocks algorithm '
+              '(default: %(default)s).'))
+    subparser.add_argument('--no-mmap',
+                           action='store_true',
+                           help='Do not use mmap.')
+    add_data_format_args(subparser)
     subparser.add_argument('fromfile', help='From file.')
     subparser.add_argument('tofile', help='To file.')
     subparser.add_argument('patchfile', help='Created patch file.')
     subparser.set_defaults(func=_do_create_patch)
 
+    # Create in-place patch subparser.
+    subparser = subparsers.add_parser('create_patch_in_place',
+                                      description='Create an in-place patch.')
+    subparser.add_argument('-c', '--compression',
+                           choices=sorted(_COMPRESSIONS),
+                           default='lzma',
+                           help='Compression algorithm (default: %(default)s).')
+    subparser.add_argument('-s', '--suffix-array-algorithm',
+                           choices=('sais', 'divsufsort'),
+                           default='divsufsort',
+                           help='Suffix array algorithm (default: %(default)s).')
+    subparser.add_argument('--memory-size',
+                           required=True,
+                           type=to_binary_size,
+                           help='Target memory size.')
+    subparser.add_argument(
+        '--segment-size',
+        required=True,
+        type=to_binary_size,
+        help='Segment size. Must be a multiple of the largest erase block size.')
+    subparser.add_argument(
+        '--minimum-shift-size',
+        type=to_binary_size,
+        help='Minimum shift size (default: 2 * segment size).')
+    subparser.add_argument('--no-mmap',
+                           action='store_true',
+                           help='Do not use mmap.')
+    add_data_format_args(subparser)
+    subparser.add_argument('fromfile', help='From file.')
+    subparser.add_argument('tofile', help='To file.')
+    subparser.add_argument('patchfile', help='Created patch file.')
+    subparser.set_defaults(func=_do_create_patch_in_place)
+
+    # Create bsdiff patch subparser.
+    subparser = subparsers.add_parser('create_patch_bsdiff',
+                                      description='Create a bsdiff patch.')
+    subparser.add_argument('fromfile', help='From file.')
+    subparser.add_argument('tofile', help='To file.')
+    subparser.add_argument('patchfile', help='Created patch file.')
+    subparser.set_defaults(func=_do_create_patch_bsdiff)
+
     # Apply patch subparser.
-    subparser = subparsers.add_parser('apply_patch',
-                                      description='Apply given patch.')
+    subparser = subparsers.add_parser(
+        'apply_patch',
+        description='Apply given sequential or hdiffpatch patch.')
     subparser.add_argument('fromfile', help='From file.')
     subparser.add_argument('patchfile', help='Patch file.')
     subparser.add_argument('tofile', help='Created to file.')
@@ -538,6 +665,9 @@ def _main():
 
     args = parser.parse_args()
 
+    level = logging.getLevelName(args.log_level.upper())
+    logging.basicConfig(level=level, format='%(asctime)s %(message)s')
+
     if args.debug:
         args.func(args)
     else:
